@@ -41,11 +41,12 @@ import sys
 from dataclasses import dataclass, field
 from collections import defaultdict
 from typing import List, Tuple
-
-
-# Exceptions for dealing with argument checking.
-class PoscBaseException(Exception):
-    """Base exception functionality"""
+from pythonopenscad.m3dapi import M3dRenderer, RenderContext, RenderContextCrossSection
+from pythonopenscad.modifier import (
+    PoscBaseException,
+    PoscRendererBase,
+    get_fragments_from_fn_fa_fs
+)
 
 
 class ConversionException(PoscBaseException):
@@ -69,10 +70,6 @@ class RequiredParameterNotProvided(PoscBaseException):
     """Exception when a required parameter is not provided."""
 
 
-class InvalidModifier(PoscBaseException):
-    """Attempting to add or remove an unknown modifier."""
-
-
 class InitializerNotAllowed(PoscBaseException):
     """An initializer (def __init__) was defined and is not allowed."""
 
@@ -416,116 +413,6 @@ class OpenScadApiSpecifier(object):
         return 'No arguments allowed.'
 
 
-class PoscMetadataBase(object):
-    """Provides medatadata properties."""
-
-    def getMetadataName(self):
-        if not hasattr(self, '_metabase_name'):
-            return ''
-        return self._metabase_name
-
-    def setMetadataName(self, value):
-        self._metabase_name = value
-        return self
-
-
-class OscModifier(object):
-    """Defines an OpenScad modifier"""
-
-    def __init__(self, modifier, name):
-        self.modifier = modifier
-        self.name = name
-
-    def __repr__(self):
-        return self.name
-
-
-DISABLE = OscModifier('*', 'DISABLE')
-SHOW_ONLY = OscModifier('!', 'SHOW_ONLY')
-DEBUG = OscModifier('#', 'DEBUG')
-TRANSPARENT = OscModifier('%', 'TRANSPARENT')
-BASE_MODIFIERS = (DISABLE, SHOW_ONLY, DEBUG, TRANSPARENT)
-BASE_MODIFIERS_SET = set(BASE_MODIFIERS)
-
-
-class PoscModifiers(PoscMetadataBase):
-    """Functions to add/remove OpenScad modifiers.
-
-    The add_modifier and remove_modifier functions can be chained as they return self.
-
-    e.g.
-    Cylinder() - Cube().add_modifier(SHOW_ONLY, DEBUG).color('#f00')
-
-    Will create a red 1x1x1 cube with the ! and # OpenScad modifiers. The SHOW_ONLY
-    modifier will cause the cylinder to not be displayed.
-        difference() {
-          cylinder(h=1.0, r=1.0, center=false);
-          !#cube(size=[1.0, 1.0, 1.0]);
-        }
-
-    This API is specified to PythonOpenScad. OpenPyScad and SolidPython use different
-    APIs for this feature.
-    """
-
-    def check_is_valid_modifier(self, *modifiers):
-        if set(modifiers) - BASE_MODIFIERS_SET:
-            raise InvalidModifier(
-                '"%r" is not a valid modifier. Muse be one of %r' % (modifiers, BASE_MODIFIERS)
-            )
-
-    def add_modifier(self, modifier, *args):
-        """Adds one of the model modifiers like DISABLE, SHOW_ONLY, DEBUG or TRANSPARENT.
-        Args:
-          modifer, *args: The modifier/a being added. Checked for validity.
-        """
-        self.check_is_valid_modifier(modifier, *args)
-        if not hasattr(self, '_osc_modifier'):
-            self._osc_modifier = set((modifier,))
-        self._osc_modifier.update(args + (modifier,))
-        return self
-
-    def remove_modifier(self, modifier, *args):
-        """Removes a modifiers, one of DISABLE, SHOW_ONLY, DEBUG or TRANSPARENT.
-        Args:
-          modifer, *args: The modifier/s being removed. Checked for validity.
-        """
-        self.check_is_valid_modifier(modifier, *args)
-        if not hasattr(self, '_osc_modifier'):
-            return
-        self._osc_modifier.difference_update(args + (modifier,))
-        return self
-
-    def has_modifier(self, modifier):
-        """Checks for presence of a modifier, one of DISABLE, SHOW_ONLY, DEBUG or TRANSPARENT.
-        Args:
-          modifer: The modifier being inspected. Checked for validity.
-        """
-        self.check_is_valid_modifier(modifier)
-        if not hasattr(self, '_osc_modifier'):
-            return False
-        return modifier in self._osc_modifier
-
-    def get_modifiers(self):
-        """Returns the current set of modifiers as an OpenScad equivalent modifier string"""
-        if not hasattr(self, '_osc_modifier'):
-            return ''
-        # Maintains order of modifiers.
-        return ''.join(i.modifier for i in BASE_MODIFIERS if i in self._osc_modifier)
-
-    def get_modifiers_repr(self):
-        """Returns the repr() equivalent of the current set or None if none are set."""
-        if not hasattr(self, '_osc_modifier'):
-            return None
-        if self._osc_modifier:
-            return repr(self._osc_modifier)
-        return None
-
-    # Deprecated.
-    def transparent(self):
-        self.add_modifier(TRANSPARENT)
-        return self
-
-
 class StringWriter(object):
     """A CodeDumper writer that writes to a string. This can API can be implemented for
     file writers or other uses."""
@@ -799,17 +686,20 @@ class CodeDumperForPython(CodeDumper):
         return ('', '.add_modifier(*%s)' % s)
 
 
-class PoscBase(PoscModifiers):
+class PoscBase(PoscRendererBase):
     DUMP_CONTAINER = True
     DUMP_MODULE = False
 
     def __post_init__(self):
         for arg in self.OSC_API_SPEC.args:
             value = getattr(self, arg.name)
+            is_different_name = arg.name != arg.attr_name
+            if is_different_name:
+                delattr(self, arg.name)
             if value is not None:
-                if arg.name != arg.attr_name:
-                    delattr(self, arg.name)
                 setattr(self, arg.attr_name, arg.typ(value))
+            elif is_different_name:
+                setattr(self, arg.attr_name, None)
 
         self.init_children()
         # Object should be fully constructed now.
@@ -841,10 +731,6 @@ class PoscBase(PoscModifiers):
                 result.append(code_dumper.render_name_value(arg, v))
         return result
 
-    def can_have_children(self):
-        """This is a childless node, always returns False."""
-        return False
-
     def has_children(self):
         return False
 
@@ -949,6 +835,9 @@ class PoscBase(PoscModifiers):
         """
         with open(filename, 'w', encoding=encoding) as fp:
             self.dump(fp)
+            
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        assert False, "Not implemented"
 
     def module(self, name):
         """Returns a variable that references this object."""
@@ -1080,7 +969,7 @@ class PoscParentBase(PoscBase):
         """Returns true if the node has children."""
         return bool(self._children)
 
-    def children(self):
+    def children(self) -> list[PoscBase]:
         """Returns the list of children"""
         return self._children
 
@@ -1156,6 +1045,9 @@ class Translate(PoscParentBase):
         (Arg('v', VECTOR3_FLOAT, None, '(x,y,z) translation vector.', required=True),),
         OPEN_SCAD_URL_TAIL_TRANSFORMS,
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.translate(self, self.v)
 
 
 @apply_posc_transformation_attributes
@@ -1178,6 +1070,9 @@ class Rotate(PoscParentBase):
         ),
         OPEN_SCAD_URL_TAIL_TRANSFORMS,
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.rotate(self, self.a, self.v)
 
 
 @apply_posc_attributes
@@ -1244,6 +1139,13 @@ class Cylinder(PoscBase):
             if v is None:
                 raise RequiredParameterNotProvided('"%s" is required and not provided' % k)
         self.check_required_parameters()
+        
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        centre = True if self.center else False
+        return renderer.cylinder(
+            self, self.h, self.get_r1(), self.get_r2(), 
+            get_fragments_from_fn_fa_fs(self.get_r1(), self._fn, self._fa, self._fs),
+            centre)
 
 
 @apply_posc_attributes
@@ -1278,6 +1180,9 @@ class Sphere(PoscBase):
                 'Both parameters r and d are None. A value for r or d must be provided.'
             )
         self.check_required_parameters()
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.sphere(self, self.r, get_fragments_from_fn_fa_fs(self.r, self._fn, self._fa, self._fs))
 
 
 @apply_posc_attributes
@@ -1300,6 +1205,9 @@ class Cube(PoscBase):
         ),
         OPEN_SCAD_URL_TAIL_PRIMITIVES,
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.cube(self, self.size, True if self.center else False)
 
 
 @apply_posc_transformation_attributes
@@ -1315,6 +1223,9 @@ class Scale(PoscParentBase):
         ),
         OPEN_SCAD_URL_TAIL_TRANSFORMS,
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.scale(self, self.v)
 
 
 @apply_posc_transformation_attributes
@@ -1340,6 +1251,9 @@ class Resize(PoscParentBase):
         ),
         OPEN_SCAD_URL_TAIL_TRANSFORMS,
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.resize(self, self.newsize, self.auto)
 
 
 @apply_posc_transformation_attributes
@@ -1351,6 +1265,9 @@ class Mirror(PoscParentBase):
         (Arg('v', VECTOR3_FLOAT, None, 'The normal of the plane to be mirrored.'),),
         OPEN_SCAD_URL_TAIL_TRANSFORMS,
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.mirror(self, self.v)
 
 
 @apply_posc_transformation_attributes
@@ -1380,6 +1297,9 @@ class Multmatrix(PoscParentBase):
     def get_m(self):
         """Returns the matrix m. The returned value is always a 4x4 matrix."""
         return self.m
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.multmatrix(self, self.m)
 
 
 @apply_posc_transformation_attributes
@@ -1401,6 +1321,9 @@ class Color(PoscParentBase):
         ),
         OPEN_SCAD_URL_TAIL_TRANSFORMS,
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.color(self, self.c, self.alpha)
 
 
 @apply_posc_transformation_attributes
@@ -1426,6 +1349,9 @@ class Offset(PoscParentBase):
         if all(x is None for x in [self.r, self.delta]):
             self.r = 1.0
         self.check_required_parameters()
+        
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.offset(self, self.r, self.delta, self.chamfer, self._fn, self._fa, self._fs)
 
 
 @apply_posc_transformation_attributes
@@ -1446,6 +1372,9 @@ class Projection(PoscParentBase):
         '3D_to_2D_Projection',
     )
     
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.projection(self, self.cut)
+    
 @apply_posc_transformation_attributes
 class Render(PoscParentBase):
     """Forces the generation of a mesh even in preview mode."""
@@ -1463,6 +1392,9 @@ class Render(PoscParentBase):
         OPEN_SCAD_URL_TAIL_OTHER,
         'render',
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.render(self, self.convexity)
 
 
 @apply_posc_transformation_attributes
@@ -1470,6 +1402,9 @@ class Minkowski(PoscParentBase):
     """Create a Minkowski transformed object."""
 
     OSC_API_SPEC = OpenScadApiSpecifier('minkowski', (), OPEN_SCAD_URL_TAIL_TRANSFORMS)
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.minkowski(self)
 
 
 @apply_posc_transformation_attributes
@@ -1477,6 +1412,9 @@ class Hull(PoscParentBase):
     """Create a hull of two solids."""
 
     OSC_API_SPEC = OpenScadApiSpecifier('hull', (), OPEN_SCAD_URL_TAIL_TRANSFORMS)
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.hull(self)
 
 
 @apply_posc_transformation_attributes
@@ -1520,6 +1458,10 @@ class Linear_Extrude(PoscParentBase):
         OPEN_SCAD_URL_TAIL_2D,
         'Linear_Extrude',
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.linear_extrude(self, self.height, self.center, self.convexity, 
+                                       self.twist, self.slices, self.scale_, self._fn)
 
 
 @apply_posc_transformation_attributes
@@ -1540,6 +1482,15 @@ class Rotate_Extrude(PoscParentBase):
         OPEN_SCAD_URL_TAIL_2D,
         'Rotate_Extrude',
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+
+        return renderer.rotate_extrude(self, 
+                                       self.angle, 
+                                       self.convexity, 
+                                       self._fn, 
+                                       self._fa, 
+                                       self._fs)
 
 
 @apply_posc_attributes
@@ -1573,6 +1524,14 @@ class Circle(PoscBase):
         if self.get_r() is None:
             raise RequiredParameterNotProvided('r or d is required and not provided')
         self.check_required_parameters()
+        
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        if self.d is None:
+            r = self.r
+        else:
+            r = self.d / 2  
+        fn = get_fragments_from_fn_fa_fs(r, self._fn, self._fa, self._fs)
+        return renderer.circle(self, r, fn)
 
 
 @apply_posc_attributes
@@ -1598,6 +1557,9 @@ class Square(PoscBase):
         OPEN_SCAD_URL_TAIL_2D,
     )
 
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.square(self, self.size, self.center)
+
 
 @apply_posc_attributes
 class Polygon(PoscBase):
@@ -1628,6 +1590,9 @@ class Polygon(PoscBase):
         OPEN_SCAD_URL_TAIL_2D,
         'Polygons',
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.polygon(self, self.points, self.paths, self.convexity)
 
 
 @apply_posc_attributes
@@ -1654,6 +1619,31 @@ class Text(PoscBase):
         'Text',
         '',
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        direction = self.direction if self.direction else "ltr"
+        size = self.size if self.size else 1.0
+        font = self.font if self.font else "Liberation Sans"
+        halign = self.halign if self.halign else "left"
+        valign = self.valign if self.valign else "baseline"
+        spacing = self.spacing if self.spacing else 1.0
+        language = self.language if self.language else "en"
+        script = self.script if self.script else "latin"
+        
+        return renderer.text(
+            self,
+            self.text, 
+            size, 
+            font, 
+            halign, 
+            valign, 
+            spacing, 
+            direction,
+            language, 
+            script, 
+            self._fa, 
+            self._fs, 
+            self._fn)
 
 
 @apply_posc_attributes
@@ -1687,6 +1677,8 @@ class Polyhedron(PoscBase):
         OPEN_SCAD_URL_TAIL_PRIMITIVES,
     )
 
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.polyhedron(self, self.points, self.triangles, self.faces)
 
 @apply_posc_attributes
 class Union(PoscParentBase):
@@ -1695,6 +1687,8 @@ class Union(PoscParentBase):
 
     OSC_API_SPEC = OpenScadApiSpecifier('union', (), OPEN_SCAD_URL_TAIL_CSG)
 
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.union(self)
 
 @apply_posc_attributes
 class LazyUnion(PoscParentBase):
@@ -1703,6 +1697,9 @@ class LazyUnion(PoscParentBase):
 
     DUMP_CONTAINER = False  # When rendering to OpenScad, don't render the container.
     OSC_API_SPEC = OpenScadApiSpecifier('lazy_union', (), OPEN_SCAD_URL_TAIL_CSG)
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.union(self)
 
 
 @apply_posc_attributes
@@ -1745,6 +1742,8 @@ class Module(PoscParentBase):
         for child in self.children():
             child.code_dump(code_dumper)
 
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.union(self)
 
 @apply_posc_attributes
 class Difference(PoscParentBase):
@@ -1753,13 +1752,18 @@ class Difference(PoscParentBase):
 
     OSC_API_SPEC = OpenScadApiSpecifier('difference', (), OPEN_SCAD_URL_TAIL_CSG)
 
-
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.difference(self)
+    
 @apply_posc_attributes
 class Intersection(PoscParentBase):
     """Creates a 3D object by finding the common space contained in all the provided
     3D objects."""
 
     OSC_API_SPEC = OpenScadApiSpecifier('intersection', (), OPEN_SCAD_URL_TAIL_CSG)
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.intersection(self)
 
 
 @apply_posc_attributes
@@ -1789,7 +1793,9 @@ class Import(PoscBase):
         ),
         OPEN_SCAD_URL_TAIL_IMPORTING,
     )
-
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.import_file(self, self.file, self.layer, self.convexity)
 
 @apply_posc_attributes
 class Surface(PoscBase):
@@ -1818,6 +1824,9 @@ class Surface(PoscBase):
         OPEN_SCAD_URL_TAIL_OTHER,
         'Surface',
     )
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.surface(self, self.file, self.center, self.invert, self.convexity)
 
 
 @apply_posc_attributes
@@ -1826,3 +1835,6 @@ class Fill(PoscParentBase):
     the result is identical to hull()."""
 
     OSC_API_SPEC = OpenScadApiSpecifier('fill', (), OPEN_SCAD_URL_TAIL_TRANSFORMS)
+    
+    def renderObj(self, renderer: M3dRenderer) -> RenderContext:
+        return renderer.fill(self)
